/*


	Plug In PYX
	Copyright (c) 2002-2022 Alchemy Mindworks


*/

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "gwsplpyx.h"

#include "plugin.h"

HINSTANCE hInst=NULL;
HINSTANCE hCallingInstance=NULL;

//this calculates the correctly-padded stride for a Windows device-independant bitmap.
#define WIDTHBYTES(i)   ((i+31)/32*4)

#define	PYX_SIGSIZE	12
#define	PYX_SIG		"PYX_graphic_"

#ifdef _WIN64
#pragma pack(push)
#pragma pack(1)
#endif
typedef struct {
	char sig[PYX_SIGSIZE];
        long x,y;
        long bits;
	char palette[768];
	} PYX_HEADER;

#ifdef _WIN64
#pragma pack(pop)
#endif

static void LoadConfig()
{
	/*
        Load any configuration items here. The example plugin doesn't have any.

        Configuration items can be saved either to the Windows registry or to an INI
        file, as Graphic Workshop itself does. The latter, while considered to be
        primitive and retro in some circles, is arguably safer and easier for users
        to edit if the need to do so arises.
        */
}

static void SaveConfig()
{
	/*
        Save any configuration items here.
        */
}

/*

	The DllMain or DllEntryPoint function is called by Windows when a DLL is loaded, and again when it's freed.
	This is a good place to fetch any configuration options pertinent to your plugin, as illustrated here.

*/

#ifdef _MSC_BUILD
//Visual C
DLL_EXPORT (BOOL) DllMain(HINSTANCE hInstance,DWORD dwReason,LPVOID lpCmdLine)
#else
//other compilers
DLL_EXPORT (BOOL) DllEntryPoint(HINSTANCE hInstance,DWORD dwReason,LPVOID lpCmdLine)
#endif
{
        hInst=hInstance;

        if(dwReason == DLL_PROCESS_ATTACH) {
        	LoadConfig();
	}

	if(dwReason == DLL_PROCESS_DETACH) {

        }

	return(TRUE);
}

/*

	The EXPORTEG_GetPluginParams function is called by Graphic Workshop when it enumerates and
        loads all its plugins. It fills in a GWSPLUGINPARAM object, as defined in plugin.h, so
        Grapic Workshop will know what your plugin is capable of, what to call it and what file
        extensions should be associated with it.

*/

DLL_EXPORT (int) EXPORTEG_GetPluginParams(GWSPLUGINPARAM *param,LPARAM lParam)
{
	unsigned long size=param->size;

        //save the calling instance somewhere local
        hCallingInstance=(HINSTANCE)param->hinstance;

        //initialize the GWSPLUGINPARAM object
        memset((char *)param,0,size);
        param->size=size;

        //tell Graphic Workshop what your plugin is called.
	//note that param->extension can list up to 32 file extensions, like this "PYX;PYY;PYZ" and so on
        strcpy(param->extension,"PYX");
        strcpy(param->description,"(PYX) Image File Format");
	strcpy(param->copyright,"PYX plugin copyright \251 2022 " PLUGIN_CREATOR ".");

        //this is the maximum image bit depth your plugin supports.
        //it can be 1,4,8 or 24.
        //if your plugin supports alpha channels, this value is still 24.
        param->maxbits=24;
        //these flags tell graphic workshop what your plugin can do.
        //for most plugins, you'll want FT_BITMAP to indicate that your plugin works with pictures - pretty much essential.
        //the FT_SUPPORTALPHA flag should be included if your plugin handles alpha transparency.
        param->flags=FT_BITMAP | FT_SUPPORTALPHA;
        param->version=PLUGINVERSION();

	return(TRUE);
}

/*

	this macro gets a pointer to a GWSPLUGINCALLBACK object, used in reading and writing files

*/

#define PREAMBLE() {\
	if(lParam != NULL) {\
        	lpcb=(GWSPLUGINCALLBACK *)lParam;\
		if(lpcb->magic == GWSPLUGINCALLBACK_MAGIC) {\
                	(*result)=R_SUCCESS;\
                }\
        }\
        else {\
                (*result)=UNSUPPORTED_FUNCTION;\
                goto bail;\
	}\
}

/*

	The EXPORTEG_WriteImage function accepts a handle to a Windows device-independent bitmap and a path to
        the file you'd like to write it to. It returns TRUE if it's successful or FALSE if it fails. If it
        fails, the reason for the failure should be in *result - the supported results codes are defined
        in plugin.h.

        Note the discussion of device-independant bitmaps in the documentation for this SDK.

        The himage argument to EXPORTEG_WriteImage is a handle to the device-independant bitmap to be written
        to a file. The device-independant bitmap mast have a bit depth of 1,4,8,24 or 32 bits, and be
        uncompressed.

*/

DLL_EXPORT (int) EXPORTEG_WriteImage(char *path,HANDLE himage,HANDLE hcomment,int *result,LPARAM lParam)
{
	PYX_HEADER pyxhead;
        GWSPLUGINCALLBACK *lpcb=NULL;
        LPBITMAPINFOHEADER lpbi=NULL;
        FILE *fp=NULL;
	char *ps;
        long stride=0;
        int i;

        //point lpcb to the callback structure provided by Graphic Workshop
        PREAMBLE();

        //initialize the PYX header object - change this for the format your plugin supports
        memset((char *)&pyxhead,0,sizeof(PYX_HEADER));

        //open the file to write - no, we don't want to hear why gotos are evil
        if((fp=fopen(path,"wb"))==NULL) {
                (*result)=BAD_OPEN;
                goto bail;
        }

        //lock the image handle
        if((lpbi=(LPBITMAPINFOHEADER)GlobalLock(himage))==NULL) {
                (*result)=BAD_ALLOC;
                goto bail;
        }

        //install the magic in the PYX header
        memcpy(pyxhead.sig,PYX_SIG,PYX_SIGSIZE);

        //if the image to be written is paletted - indicated by it having 8 or fewer bits of color - extract its palette
        if(lpbi->biBitCount <= 8) {
        	RGBQUAD *rgb=(RGBQUAD *)((char *)lpbi+(long)lpbi->biSize);
		int j=min(1<<lpbi->biBitCount,256);

		for(i=0;i<j;i++) {
			pyxhead.palette[i*3]=rgb[i].rgbRed;
			pyxhead.palette[i*3+1]=rgb[i].rgbGreen;
			pyxhead.palette[i*3+2]=rgb[i].rgbBlue;
		}
        }

        //fill in the remaining fields of the PYX file header. your format will no doub vary.
	pyxhead.x=lpbi->biWidth;
        pyxhead.y=lpbi->biHeight;
        pyxhead.bits=lpbi->biBitCount;

        //write the file header
	if(fwrite((char *)&pyxhead,1,sizeof(PYX_HEADER),fp) != sizeof(PYX_HEADER)) {
                (*result)=BAD_WRITE;
                goto bail;
        }

        //locate the image data in the device-independant bitmap.
        //keep in mind that the data is upside down - this points to the last row of the image
        ps=((char *)lpbi+lpbi->biSize+(lpbi->biBitCount <= 8 ? (long)((1<<lpbi->biBitCount)*sizeof(RGBQUAD)) : 0));

        //calculate the width in bytes of one row of image data.
        //windows pads this value out, as illustrated by the WIDTHBYTES macro
        stride=(WIDTHBYTES((long)lpbi->biWidth*(long)lpbi->biBitCount));

        //set the text in Graphic Workshop's status field
        (lpcb->openstatus)();
        (lpcb->statustext)((char *)"Writing image");

        //step through the source rows of the image and write each one to the file
        //your format may be a lot more complex than this, especially if you have to swab the byte order of its pixels
        for(i=0;i<pyxhead.y;++i) {
                char *pss=ps+i*stride;

                if(fwrite(pss,1,stride,fp) != stride) {
	                (*result)=BAD_WRITE;
                        goto bail;
	        }

                //update to status bar graph in Graphic Workshop
                (lpcb->updatestatus)(i,pyxhead.y);
        }

        //we're done
bail:
	//blank the text in Graphic Workshop's status field
	(lpcb->statustext)((char *)"");
        (lpcb->closestatus)();
        //close the file
        if(fp != NULL) fclose(fp);
        //if the write failed, remove the damaged file
        if((*result) != R_SUCCESS) remove(path);
        //unlock the himage handle
        if(himage != NULL && lpbi != NULL) GlobalUnlock(himage);
	//return the status of the write
	return((*result) == R_SUCCESS);
}

/*

	The EXPORTEG_ReadImage function accepts a path to the file to be read, and returns
        a handle to a Windows device-independant bitmap that contains the contents of the
        read image. It returns NULL if the read was unsuccessful. If it fails, the reason
        for the failure should be in *result - the supported results codes are defined
        in plugin.h.

*/

DLL_EXPORT (HANDLE) EXPORTEG_ReadImage(char *path,int *result,LPARAM lParam)
{
	FILE *fp=NULL;
	PYX_HEADER pyxhead;
	GWSPLUGINCALLBACK *lpcb=NULL;
	HANDLE himage=NULL;
        LPBITMAPINFOHEADER lpbi=NULL;
        char *pd;
        long imagesize,colorsused;
	int i,j,stride;

        //point lpcb to the callback structure provided by Graphic Workshop
        PREAMBLE();

        //open the file to read
        if((fp=fopen(path,"rb"))==NULL) {
                (*result)=BAD_OPEN;
                goto bail;
        }

        //read the PYX header
        if(fread((char *)&pyxhead,1,sizeof(PYX_HEADER),fp) != sizeof(PYX_HEADER)) {
                (*result)=BAD_READ;
                goto bail;
        }

        //check to make sure it really is a PYX file - this will be different for your format.
        if(memcmp(pyxhead.sig,PYX_SIG,PYX_SIGSIZE)) {
                (*result)=BAD_FILE;
                goto bail;
        }

        //calculate the size of the image data in the file to be read based on the dimensions stored in the header.
        imagesize=WIDTHBYTES(pyxhead.x*pyxhead.bits)*pyxhead.y;
        //calculate the number colors used if the image in the file being read is paletted, that is, if it has 8 or fewer bits of color.
        colorsused=((pyxhead.bits <= 8) ? ((1<<pyxhead.bits) & 0xffff) : 0);

        //allocate a device-independant bitmap
	if((himage=GlobalAlloc(GMEM_MOVEABLE | GMEM_ZEROINIT,sizeof(BITMAPINFOHEADER)+colorsused*sizeof(RGBQUAD)+imagesize))==NULL) {
        	(*result)=BAD_ALLOC;
                goto bail;
	}

        //lock the device-independant bitmap
        if((lpbi=(LPBITMAPINFOHEADER)GlobalLock(himage))==NULL) {
                (*result)=BAD_ALLOC;
                goto bail;
        }

        //complete the BITMAPINFOHEADER fields of the newly-allocated device-independant bitmap
        lpbi->biSize=sizeof(BITMAPINFOHEADER);
        lpbi->biPlanes=1;
	lpbi->biBitCount=pyxhead.bits;
        lpbi->biWidth=pyxhead.x;
        lpbi->biHeight=pyxhead.y;
        lpbi->biCompression=BI_RGB;
	lpbi->biSizeImage=0;
        lpbi->biXPelsPerMeter=0;
        lpbi->biYPelsPerMeter=0;
	lpbi->biClrUsed=colorsused;
	lpbi->biClrImportant=lpbi->biClrUsed;

        //add an array of RGBQUAD aobjects to the device-independant bitmap if this image is paletted
        if(pyxhead.bits <= 8) {
		RGBQUAD *pRgb=(RGBQUAD *)((char *)lpbi+(unsigned long)lpbi->biSize);

		for(i=0;i<colorsused;++i) {
			pRgb[i].rgbRed=pyxhead.palette[i*3];
			pRgb[i].rgbGreen=pyxhead.palette[i*3+1];
			pRgb[i].rgbBlue=pyxhead.palette[i*3+2];
			pRgb[i].rgbReserved=0;
		}
        }

        //locate the image data in the device-independant bitmap.
	pd=((char *)lpbi+lpbi->biSize+(lpbi->biBitCount <= 8 ? (long)((1<<lpbi->biBitCount)*sizeof(RGBQUAD)) : 0));

        //calculate the width in bytes of one row of image data.
        //windows pads this value out, as illustrated by the WIDTHBYTES macro
        stride=(WIDTHBYTES((long)lpbi->biWidth*(long)lpbi->biBitCount));

        //set the text in Graphic Workshop's status field
        (lpcb->openstatus)();
        (lpcb->statustext)((char *)"Reading image");

        //step through the source rows of the image and read each one from the file
        for(i=0;i<pyxhead.y;++i) {
                char *pdd=pd+i*stride;

                if(fread(pdd,1,stride,fp) != stride) {
	                (*result)=BAD_WRITE;
        	        goto bail;
	        }

                //update to status bar graph in Graphic Workshop
                (lpcb->updatestatus)(i,pyxhead.y);
        }

        //we're done
bail:
	//blank the text in Graphic Workshop's status field
	(lpcb->statustext)((char *)"");
        (lpcb->closestatus)();
        //unlock the device-independant bitmap
	if(himage != NULL && lpbi != NULL) GlobalUnlock(himage);
        //if the read was unsuccessful, free the device-independant bitmap
        if(himage != NULL && (*result) != R_SUCCESS) GlobalFree(himage);
        //close the file
        if(fp != NULL) fclose(fp);

        //return a handle to the device-independant bitmap or NULL if the read failed
	return((*result) == R_SUCCESS ? himage : NULL);
}

/*

	The EXPORTEG_GetInfo reads enough of the file about to be opened to determine its
        pixel dimensions, file size and bit depth, and completes a GWSPLUGINFILEINFO with
        these values. It returns TRUE if it's successful or FALSE if it fails. If it fails,
        the reason for the failure should be in *result - the supported results codes are
        defined in plugin.h.

*/

DLL_EXPORT (int) EXPORTEG_GetInfo(GWSPLUGINFILEINFO *info,char *path,int *result,LPARAM lParam)
{
        PYX_HEADER pyxhead;
        FILE *fp=NULL;
        HANDLE himage=NULL;
        LPBITMAPINFOHEADER lpbi=NULL;
        long size=info->size;
        long filesize;

        //initilize the GWSPLUGINFILEINFO object
        memset((char *)info,0,size);

        info->size=size;
        info->version=PLUGINVERSION();

        //open the file for reading
        if((fp=fopen(path,"rb"))==NULL) {
                (*result)=BAD_OPEN;
                goto bail;
        }

        //determine the file size
        fseek(fp,0L,SEEK_END);
        filesize=ftell(fp);
        fseek(fp,0L,SEEK_SET);

        //read the PYX header
        if(fread((char *)&pyxhead,1,sizeof(PYX_HEADER),fp) != sizeof(PYX_HEADER)) {
                (*result)=BAD_READ;
                goto bail;
        }

        //check to make sure it really is a PYX file - this will be different for your format.
        if(memcmp(pyxhead.sig,PYX_SIG,PYX_SIGSIZE)) {
                (*result)=BAD_FILE;
                goto bail;
        }

        //complete the GWSPLUGINFILEINFO object using the values from the recently read PYX header
        info->width=pyxhead.x;
        info->depth=pyxhead.y;
        info->bits=pyxhead.bits;
        info->imagesize=WIDTHBYTES(info->width*info->bits)*info->depth;
        info->filesize=filesize;
        memcpy(info->palette,pyxhead.palette,768);

        //we're done
bail:
	//close the file
	if(fp != NULL) fclose(fp);

        return((*result)==R_SUCCESS);
}

#define	SETUP_CONFIGURATION_TEXT	"This is an example Graphic Workshop plugin configuration window. It will appear if someone clicks on the Go button beside the plugin list in Graphic Workshop's Setup window.\n\n"\
					"In a real-world plugin, you'd probably want configurable items included herein. You might also have this window assert any copyright information that pertains to your plugin.\n\n"\
					"You can include a Help button in this window to access the documentation for your plugin if needs be."

/*

	This is the dialog function for EXPORTEG_DoConfigure, below. In a more complex plugin, it would
        allow users toset configurable items - this one only displays the above notice.

*/

static LRESULT FAR PASCAL PLUGINDoConfigureProc(HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam)
{
	static GWSPLUGINCONFIGPARAM *cp;
        RECT rect;
        int x,y;

	switch(message) {
		case WM_INITDIALOG:
                	//get a pointer to the GWSPLUGINCONFIGPARAM object passed to EXPORTEG_DoConfigure
                        //this is useful for more complex configuration dialogs
                	cp=(GWSPLUGINCONFIGPARAM *)lParam;

                        //set the text message in the SETUP_MESSAGE control
                        SetDlgItemText(hwnd,SETUP_MESSAGE,SETUP_CONFIGURATION_TEXT);
                        //make the SETUP_MESSAGE display a bold font
                        SendDlgItemMessage(hwnd,SETUP_MESSAGE,WM_SETFONT,(WPARAM)GetStockObject(SYSTEM_FONT),0);

                        //center the dialog window on your desktop
                        GetWindowRect(hwnd,&rect);
        		x=(GetSystemMetrics(SM_CXSCREEN)-(rect.right-rect.left))/2;
        		y=(GetSystemMetrics(SM_CYSCREEN)-(rect.bottom-rect.top))/2;
        		SetWindowPos(hwnd,NULL,x,y,rect.right-rect.left,rect.bottom-rect.top,SWP_NOSIZE);

			return(FALSE);

		case WM_COMMAND:
			switch(LOWORD(wParam)) {
			   	case IDOK:
					EndDialog(hwnd,TRUE);
					return(FALSE);
			}
			break;
	}

	return(FALSE);
}

/*
	this function displays a configuration dialog for the plugin. The plugin has the responsibility
        of storing its configuration data, if any. This function must exist, even if it only displays a
        copyright notice. lParam points to a GWSPLUGINCONFIGPARAM object.

        this function gets called when a user of Graphic Workshop selects your plugin from the plugin
        list in Graphic Workshop's Setup window and then clicks on Go.
*/

DLL_EXPORT (int) EXPORTEG_DoConfigure(HWND hwnd,LPARAM lParam)
{
	//open the configuration dialog
	int r=DialogBoxParam(hInst,MAKEINTRESOURCE(1000),hwnd,(DLGPROC)PLUGINDoConfigureProc,lParam);

        //save the updated configuration fields if the user clicked on OK - assuming that there are
        //any configurable items.
        if(r) SaveConfig();

	UpdateWindow(hwnd);
	return(TRUE);
}

